#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-07
# @Filename: acquisition.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
import time
import warnings
from dataclasses import dataclass, field
from functools import partial

from typing import TYPE_CHECKING, Any, Coroutine

import numpy
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS, FITSFixedWarning

from clu.command import FakeCommand

from cherno import config, log
from cherno.astrometry import AstrometryNet
from cherno.coordinates import gfa_to_radec
from cherno.exceptions import ChernoError
from cherno.extraction import Extraction, ExtractionData, PathLike
from cherno.maskbits import GuiderStatus
from cherno.tcc import apply_axes_correction, apply_focus_correction
from cherno.utils import astrometry_fit, focus_fit, run_in_executor


if TYPE_CHECKING:
    from cherno.actor import ChernoCommandType
    from cherno.actor.actor import ChernoState


warnings.filterwarnings("ignore", module="astropy.wcs.wcs")
warnings.filterwarnings("ignore", category=FITSFixedWarning)


__all__ = ["Acquisition"]


@dataclass
class AcquisitionData:
    """Data from the acquisition process."""

    extraction_data: ExtractionData
    solved: bool = False
    solve_time: float = -999.0
    wcs: WCS = field(default_factory=WCS)
    camera_racen: float = -999.0
    camera_deccen: float = -999.0
    xrot: float = -999.0
    yrot: float = -999.0
    rotation: float = -999.0
    proc_image: pathlib.Path | None = None

    def __post_init__(self):

        self.path = self.extraction_data.path
        self.camera = self.extraction_data.camera
        self.exposure_no = self.extraction_data.exposure_no
        self.obstime = self.extraction_data.obstime
        self.observatory = self.extraction_data.observatory
        self.field_ra = self.extraction_data.field_ra
        self.field_dec = self.extraction_data.field_dec
        self.field_pa = self.extraction_data.field_pa

    @property
    def e_data(self):
        """Shortcut for ``extraction_data``."""

        return self.extraction_data


@dataclass
class AstrometricSolution:
    """Astrometric solution data."""

    valid_solution: bool
    acquisition_data: list[AcquisitionData]
    delta_ra: float = -999.0
    delta_dec: float = -999.0
    delta_rot: float = -999.0
    delta_scale: float = -999.0
    delta_focus: float = -999.0
    rms: float = -999.0
    fwhm_median: float = -999.0
    fwhm_fit: float = -999.0
    focus_coeff: list[float] = field(default_factory=lambda: [-999.0] * 3)
    focus_r2: float = -999.0
    camera_rotation: float = -999.0
    correction_applied: list[float] = field(default_factory=lambda: [0.0] * 5)


class Acquisition:
    """Runs the steps for field acquisition."""

    def __init__(
        self,
        observatory: str,
        extractor: Extraction | None = None,
        astrometry: AstrometryNet | None = None,
        command: ChernoCommandType | None = None,
        astrometry_params: dict = {},
        extraction_params: dict = {},
    ):

        self.extractor = extractor or Extraction(observatory, **extraction_params)

        if astrometry is not None:
            self.astrometry = astrometry
        else:
            pixel_scale = config["pixel_scale"][observatory.upper()]
            backend_config = pathlib.Path(__file__).parent / "etc/astrometrynet.cfg"
            self.astrometry = AstrometryNet(
                backend_config=str(backend_config),
                width=2048,
                height=2048,
                no_plots=True,
                scale_low=pixel_scale * 0.9,
                scale_high=pixel_scale * 1.1,
                scale_units="arcsecperpix",
                sort_column="flux",
                radius=0.5,
                cpulimit=config["acquisition"]["cpulimit"],
                **astrometry_params,
            )
        self.observatory = observatory.upper()
        self.command = command or FakeCommand(log)

    def set_command(self, command: ChernoCommandType):
        """Sets the command."""

        self.command = command

    async def process(
        self,
        command: ChernoCommandType | None,
        images: PathLike | list[PathLike],
        write_proc: bool = True,
        overwrite: bool = False,
        correct: bool = True,
        full_correction: bool = False,
        offset: list[float] | None = None,
        scale_rms: bool = False,
    ):
        """Performs extraction and astrometry."""

        if command is not None:
            self.set_command(command)

        if not isinstance(images, (list, tuple)):
            images = [images]

        self.command.info("Extracting sources.")

        ext_data = await asyncio.gather(
            *[
                run_in_executor(self.extractor.process, im, executor="process")
                for im in images
            ]
        )

        for d in ext_data:
            if d.nvalid < 5:
                self.command.warning(f"Camera {d.camera}: not enough sources.")
            else:
                self.command.info(
                    fwhm_camera=[
                        d.camera,
                        d.exposure_no,
                        d.fwhm_median,
                        d.nregions,
                        d.nvalid,
                    ]
                )

        self.command.info("Running astrometry.net.")
        acq_data = await asyncio.gather(*[self._astrometry_one(d) for d in ext_data])

        self.command.debug("Saving proc- file.")
        if write_proc:
            await asyncio.gather(
                *[self.write_proc_image(d, overwrite=overwrite) for d in acq_data]
            )

        ast_solution = await self.fit(
            list(acq_data),
            offset=offset,
            scale_rms=scale_rms,
        )

        if correct and ast_solution.valid_solution is True:
            await self.correct(ast_solution, full=full_correction)

        if self.command.actor:
            update_proc_headers(ast_solution, self.command.actor.state)

        return ast_solution

    async def fit(
        self,
        data: list[AcquisitionData],
        offset: list[float] | None = None,
        scale_rms: bool = False,
    ):
        """Calculate the astrometric solution."""

        ast_solution = AstrometricSolution(False, data)

        solved = sorted([d for d in data if d.solved is True], key=lambda x: x.camera)
        weights = [s.extraction_data.nvalid for s in solved]

        if len(solved) == 0:
            self.command.error(acquisition_valid=False, did_correct=False)

        fwhm = numpy.average([d.e_data.fwhm_median for d in solved], weights=weights)
        camera_rotation = numpy.average([d.rotation for d in solved], weights=weights)

        ast_solution.fwhm_median = numpy.round(float(fwhm), 3)
        ast_solution.camera_rotation = numpy.round(float(camera_rotation), 2)

        if solved[0].field_ra == "NaN" or isinstance(solved[0].field_ra, str):
            self.command.error(acquisition_valid=False, did_correct=False)
            self.command.error("Field not defined. Cannot run astrometric fit.")
            return ast_solution

        if offset is None:
            if self.command.actor:
                offset = list(self.command.actor.state.offset)
            else:
                offset = list(config.get("offset", [0.0, 0.0, 0.0]))

        if any(offset):
            self.command.warning(offset=offset)
        else:
            self.command.debug(offset=offset)

        ast_fit = astrometry_fit(
            solved,
            offset=offset,
            obstime=solved[0].obstime.jd,
            scale_rms=scale_rms,
        )

        exp_no = solved[0].exposure_no  # Should be the same for all.

        if ast_fit is False:
            rms = delta_ra = delta_dec = delta_rot = delta_scale = -999.0
        else:
            delta_ra = ast_fit[0]
            delta_dec = ast_fit[1]
            delta_rot = ast_fit[2]
            delta_scale = ast_fit[3]

            xrms = ast_fit[4]
            yrms = ast_fit[5]
            rms = ast_fit[6]

            self.command.info(guide_rms=[exp_no, xrms, yrms, rms])

        ast_solution.valid_solution = True
        ast_solution.delta_ra = delta_ra
        ast_solution.delta_dec = delta_dec
        ast_solution.delta_rot = delta_rot
        ast_solution.delta_scale = delta_scale
        ast_solution.rms = rms

        try:
            fwhm_fit, x_min, a, b, c, r2 = focus_fit(
                [d.e_data for d in data],
                plot=config["acquisition"]["plot_focus"],
            )

            # Relationship between M2 move and focal plane. See
            # http://www.loptics.com/ATM/mirror_making/cass_info/cass_info.html
            focus_sensitivity = config["focus_sensitivity"][self.observatory]

            ast_solution.fwhm_fit = round(fwhm_fit, 3)
            ast_solution.delta_focus = round(-x_min / focus_sensitivity, 1)
            ast_solution.focus_coeff = [a, b, c]
            ast_solution.focus_r2 = round(r2, 3)

        except Exception as err:
            self.command.warning(f"Failed fitting focus curve: {err}.")

        self.command.info(
            astrometry_fit=[
                exp_no,
                len(solved),
                -999.0,
                -999.0,
                numpy.round(fwhm, 2),
                -999,
                numpy.round(camera_rotation, 3),
                delta_ra,
                delta_dec,
                delta_rot,
                delta_scale,
            ]
        )

        # Output focus data in a single keyword, mostly for Boson's benefit.
        focus_data = [int(exp_no)]
        for d in ast_solution.acquisition_data:
            if d.extraction_data.fwhm_median > 0 and d.extraction_data.nvalid > 0:
                focus_data += [
                    int(d.camera[-1]),
                    d.extraction_data.fwhm_median,
                    d.extraction_data.focus_offset,
                ]

        self.command.debug(focus_data=focus_data)

        self.command.info(
            focus_fit=[
                exp_no,
                ast_solution.fwhm_fit,
                float(f"{ast_solution.focus_coeff[0]:.3e}"),
                float(f"{ast_solution.focus_coeff[1]:.3e}"),
                float(f"{ast_solution.focus_coeff[2]:.3e}"),
                ast_solution.focus_r2,
                ast_solution.delta_focus,
            ]
        )

        if delta_scale > 0 and self.command.actor and fwhm < 2.5:
            # If we measured the scale, add it to the actor state. This is later
            # used to compute the average scale over a period. We also add the time
            # because we'll want to reject measurements that are too old.
            self.command.actor.state.scale_history.append((time.time(), delta_scale))

        return ast_solution

    async def correct(self, data: AstrometricSolution, full: bool = False):
        """Runs the astrometric fit"""

        if not self.command.actor or self.command.status.is_done:
            raise ChernoError("Cannot run correct without a valid running command.")

        actor_state = self.command.actor.state
        guider_status = actor_state.status

        stopping = (
            guider_status & (GuiderStatus.STOPPING | GuiderStatus.IDLE)
        ).value > 0

        if stopping:
            data.correction_applied = [0.0, 0.0, 0.0, 0.0, 0.0]
            self.command.info(acquisition_valid=True, did_correct=False)
            return True

        self.command.info("Applying corrections.")

        min_isolated = actor_state.guide_loop["rot"]["min_isolated_correction"]
        if abs(data.delta_rot) >= min_isolated:
            self.command.debug("Applying only large rotator correction.")
            coro = apply_axes_correction(
                self.command,
                rot=-data.delta_rot,
                k_rot=None if full is False else 1.0,
            )

        else:
            coro = apply_axes_correction(
                self.command,
                rot=-data.delta_rot,
                radec=(-data.delta_ra, -data.delta_dec),
                k_radec=None if full is False else 1.0,
                k_rot=None if full is False else 1.0,
            )

        correct_tasks: list[Coroutine[Any, Any, Any]] = [coro]

        do_focus: bool = False

        # Ignore focus correction when the r2 correlation is bad or when we got
        # an inverted parabola.
        if (
            data.focus_r2 > config["acquisition"]["focus_r2_threshold"]
            and data.focus_coeff[0] > 0
        ):
            do_focus = True
            coro = apply_focus_correction(self.command, -data.delta_focus, k_focus=None)
            correct_tasks.append(coro)
        else:
            self.command.warning("Focus fit poorly constrained. Not correcting focus.")

        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="add")
        applied_corrections: Any = await asyncio.gather(*correct_tasks)
        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="remove")

        data.correction_applied[:4] = applied_corrections[0]
        if do_focus:
            data.correction_applied[4] = applied_corrections[1] or 0.0
        else:
            data.correction_applied[4] = 0.0

        self.command.info(
            acquisition_valid=True,
            did_correct=any(data.correction_applied),
            correction_applied=data.correction_applied,
        )

    async def _astrometry_one(self, ext_data: ExtractionData):

        regions = ext_data.regions

        if self.astrometry._options.get("dir", None) is None:
            astrometry_dir = pathlib.Path(config["acquisition"]["astrometry_dir"])
            if astrometry_dir.is_absolute():
                pass
            else:
                astrometry_dir = ext_data.path.parent / astrometry_dir

            astrometry_dir.mkdir(exist_ok=True, parents=True)
        else:
            astrometry_dir = pathlib.Path(self.astrometry._options["dir"])

        outfile_root = astrometry_dir / ext_data.path.stem

        # We use all detections, even invalid ones here.
        if ext_data.algorithm == "marginal":
            xyls_df = regions.loc[:, ["x", "y", "flux"]].copy()
            xyls_df = xyls_df.rename(columns={"x_fit": "x", "y_fit": "y"})
        elif ext_data.algorithm == "daophot":
            xyls_df = regions.loc[:, ["x_0", "y_0", "flux_0"]].copy()
            xyls_df = xyls_df.rename(columns={"x_0": "x", "y_0": "y", "flux_0": "flux"})
        else:
            xyls_df = regions.loc[:, ["x", "y", "flux"]]

        gfa_xyls = Table.from_pandas(xyls_df)
        gfa_xyls_path = outfile_root.with_suffix(".xyls")
        gfa_xyls.write(str(gfa_xyls_path), format="fits", overwrite=True)

        wcs_output = outfile_root.with_suffix(".wcs")
        wcs_output.unlink(missing_ok=True)

        camera_id = int(ext_data.camera[-1])
        radec_centre = gfa_to_radec(
            1024,
            1024,
            camera_id,
            ext_data.field_ra,
            ext_data.field_dec,
            position_angle=ext_data.field_pa,
            site_name=ext_data.observatory,
        )

        proc = await self.astrometry.run(
            [gfa_xyls_path],
            stdout=outfile_root.with_suffix(".stdout"),
            stderr=outfile_root.with_suffix(".stderr"),
            ra=radec_centre[0],
            dec=radec_centre[1],
        )

        acq_data = AcquisitionData(ext_data, solve_time=proc.elapsed)

        if wcs_output.exists():
            acq_data.solved = True
            wcs = WCS(open(wcs_output).read())
            acq_data.wcs = wcs

            racen, deccen = wcs.pixel_to_world_values([[1024, 1024]])[0]
            acq_data.camera_racen = numpy.round(racen, 6)
            acq_data.camera_deccen = numpy.round(deccen, 6)

            # TODO: consider parallactic angle here.
            cd: numpy.ndarray = wcs.wcs.cd
            rot_rad = numpy.arctan2(
                numpy.array([cd[0, 1], cd[0, 0]]),
                numpy.array([cd[1, 1], cd[1, 0]]),
            )

            # Rotation is from N to E to the x and y axes of the GFA.
            yrot, xrot = numpy.rad2deg(rot_rad)
            acq_data.xrot = numpy.round(xrot % 360.0, 3)
            acq_data.yrot = numpy.round(yrot % 360.0, 3)

            # Calculate field rotation.
            camera_rot = config["cameras"]["rotation"][acq_data.camera]
            rotation = numpy.array([xrot - camera_rot - 90, yrot - camera_rot]) % 360
            rotation[rotation > 180.0] -= 360.0
            rotation = numpy.mean(rotation)
            acq_data.rotation = numpy.round(rotation, 3)

        self.command.info(
            camera_solution=[
                acq_data.camera,
                acq_data.exposure_no,
                acq_data.solved,
                acq_data.camera_racen,
                acq_data.camera_deccen,
                acq_data.xrot,
                acq_data.yrot,
                acq_data.rotation,
            ]
        )

        return acq_data

    async def write_proc_image(
        self,
        acq_data: AcquisitionData,
        outpath: PathLike = None,
        overwrite: bool = False,
    ):
        """Writes the proc-gimg image."""

        ext_data = acq_data.extraction_data

        proc_hdu = fits.open(str(acq_data.path)).copy()

        rec = Table.from_pandas(acq_data.extraction_data.regions).as_array()
        proc_hdu.append(fits.BinTableHDU(rec, name="CENTROIDS"))

        proc_hdu[1].header["SOLVED"] = acq_data.solved
        proc_hdu[1].header["SOLVTIME"] = (
            acq_data.solve_time,
            "Time to solve the field or fail",
        )

        if self.command.actor is not None:
            offsets = self.command.actor.state.offset
        else:
            offsets = [-999.0] * 3

        proc_hdu[1].header["OFFRA"] = (offsets[0], "Relative offset in RA [arcsec]")
        proc_hdu[1].header["OFFDEC"] = (offsets[1], "Relative offset in Dec [arcsec]")
        proc_hdu[1].header["OFFPA"] = (offsets[2], "Relative offset in PA [arcsec]")

        default_offset = config.get("default_offset", (0.0, 0.0, 0.0))
        aoffset = (
            default_offset[0] + offsets[0],
            default_offset[1] + offsets[1],
            default_offset[2] + offsets[2],
        )
        proc_hdu[1].header["AOFFRA"] = (aoffset[0], "Absolute offset in RA [arcsec]")
        proc_hdu[1].header["AOFFDEC"] = (aoffset[1], "Absolute offset in Dec [arcsec]")
        proc_hdu[1].header["AOFFPA"] = (aoffset[2], "Absolute offset in PA [arcsec]")

        proc_hdu[1].header.update(acq_data.wcs.to_header())

        proc_path: pathlib.Path
        if outpath is not None:
            proc_path = pathlib.Path(outpath)
            if proc_path.is_dir():
                proc_path = proc_path / ("proc-" + ext_data.path.name)
        else:
            proc_path = ext_data.path.parent / ("proc-" + ext_data.path.name)

        loop = asyncio.get_running_loop()
        func = partial(
            proc_hdu.writeto,
            proc_path,
            overwrite=overwrite,
            output_verify="silentfix",
        )
        await loop.run_in_executor(None, func)

        acq_data.proc_image = proc_path


def update_proc_headers(data: AstrometricSolution, guider_state: ChernoState):

    guide_loop = guider_state.guide_loop

    enabled_axes = guider_state.enabled_axes
    enabled_radec = "radec" in enabled_axes
    enabled_rot = "rot" in enabled_axes
    enabled_focus = "focus" in enabled_axes

    cra, cdec, crot, cscl, cfoc = data.correction_applied

    radec_pid_k = guide_loop["radec"]["pid"]["k"]
    radec_pid_td = guide_loop["radec"]["pid"].get("Td", 0.0)
    radec_pid_ti = guide_loop["radec"]["pid"].get("Ti", 0.0)

    rot_pid_k = guide_loop["rot"]["pid"]["k"]
    rot_pid_td = guide_loop["rot"]["pid"].get("Td", 0.0)
    rot_pid_ti = guide_loop["rot"]["pid"].get("Ti", 0.0)

    focus_pid_k = guide_loop["focus"]["pid"]["k"]
    focus_pid_td = guide_loop["focus"]["pid"].get("Td", 0.0)
    focus_pid_ti = guide_loop["focus"]["pid"].get("Ti", 0.0)

    if "scale" in guide_loop:
        scale_pid_k = guide_loop["scale"]["pid"]["k"]
        scale_pid_td = guide_loop["scale"]["pid"].get("Td", 0.0)
        scale_pid_ti = guide_loop["scale"]["pid"].get("Ti", 0.0)
    else:
        scale_pid_k = 0.0
        scale_pid_td = 0.0
        scale_pid_ti = 0.0

    rms = data.rms
    delta_ra = data.delta_ra
    delta_dec = data.delta_dec
    delta_rot = data.delta_rot
    delta_scale = data.delta_scale
    delta_focus = data.delta_focus

    # Update headers of proc images with deltas.
    for a_data in data.acquisition_data:
        if a_data.proc_image is not None:
            hdus = fits.open(str(a_data.proc_image), mode="update")
            header = hdus[1].header

            header["EXTMETH"] = (a_data.e_data.algorithm, "Algorithm for star finding")
            header["RADECK"] = (radec_pid_k, "PID K term for RA/Dec")
            header["RADECTD"] = (radec_pid_td, "PID Td term for RA/Dec")
            header["RADECTI"] = (radec_pid_ti, "PID Ti term for RA/Dec")

            header["ROTK"] = (rot_pid_k, "PID K term for Rot.")
            header["ROTTD"] = (rot_pid_td, "PID Td term for Rot.")
            header["ROTTI"] = (rot_pid_ti, "PID Ti term for Rot.")

            header["SCLK"] = (scale_pid_k, "PID K term for Scale")
            header["SCLTD"] = (scale_pid_td, "PID Td term for Scale")
            header["SCLTI"] = (scale_pid_ti, "PID Ti term for Scale")

            header["FOCUSK"] = (focus_pid_k, "PID K term for Focus")
            header["FOCUSTD"] = (focus_pid_td, "PID Td term for Focus")
            header["FOCUSTI"] = (focus_pid_ti, "PID Ti term for Focus")

            header["FWHM"] = (a_data.e_data.fwhm_median, "Mesured FWHM [arcsec]")
            header["FWHMFIT"] = (data.fwhm_fit, "Fitted FWHM [arcsec]")

            header["RMS"] = (rms, "Guide RMS [arcsec]")

            header["E_RADEC"] = (enabled_radec, "RA/Dec corrections enabled?")
            header["E_ROT"] = (enabled_rot, "Rotator corrections enabled?")
            header["E_FOCUS"] = (enabled_focus, "Focus corrections enabled?")
            header["E_SCL"] = (False, "Scale corrections enabled?")

            header["DELTARA"] = (delta_ra, "RA measured delta [arcsec]")
            header["DELTADEC"] = (delta_dec, "Dec measured delta [arcsec]")
            header["DELTAROT"] = (delta_rot, "Rotator measured delta [arcsec]")
            header["DELTASCL"] = (delta_scale, "Scale measured factor")
            header["DELTAFOC"] = (delta_focus, "Focus delta [microns]")

            header["CORR_RA"] = (cra, "RA applied correction [arcsec]")
            header["CORR_DEC"] = (cdec, "Dec applied correction [arcsec]")
            header["CORR_ROT"] = (crot, "Rotator applied correction [arcsec]")
            header["CORR_SCL"] = (cscl, "Scale applied correction")
            header["CORR_FOC"] = (cfoc, "Focus applied correction [microns]")

            hdus.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-10
# @Filename: coordinates.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pandas
from coordio import ICRS, Field, FocalPlane, Observed, Site, Wok
from coordio.conv import guideToTangent, tangentToWok
from coordio.defaults import INST_TO_WAVE, calibration


def gfa_to_wok(xPix: float, yPix: float, gfaID: int):
    """Converts from a GFA pixel to wok coordinates."""

    idx = pandas.IndexSlice

    gfaRow = calibration.gfaCoords.loc[idx[:, gfaID], :]

    b = gfaRow[["xWok", "yWok", "zWok"]].to_numpy().squeeze()
    iHat = gfaRow[["ix", "iy", "iz"]].to_numpy().squeeze()
    jHat = gfaRow[["jx", "jy", "jz"]].to_numpy().squeeze()
    kHat = gfaRow[["kx", "ky", "kz"]].to_numpy().squeeze()

    xt, yt = guideToTangent(xPix, yPix)
    zt = 0

    return tangentToWok(xt, yt, zt, b, iHat, jHat, kHat)  # type: ignore


def gfa_to_radec(
    xPix: float,
    yPix: float,
    gfaID: int,
    bore_ra: float,
    bore_dec: float,
    position_angle: float = 0,
    site_name: str = "APO",
):
    """Converts from a GFA pixel to observed RA/Dec."""

    site = Site(site_name)
    site.set_time()

    wavelength = INST_TO_WAVE["GFA"]

    wok_coords = gfa_to_wok(xPix, yPix, gfaID)

    boresight_icrs = ICRS([[bore_ra, bore_dec]])
    boresight = Observed(
        boresight_icrs,
        site=site,
        wavelength=wavelength,
    )

    wok = Wok([wok_coords], site=site, obsAngle=position_angle)
    focal = FocalPlane(wok, wavelength=wavelength, site=site)
    field = Field(focal, field_center=boresight)
    observed = Observed(field, wavelength=wavelength, site=site)

    return (observed.ra[0], observed.dec[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-12
# @Filename: tcc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy

from cherno.exceptions import ChernoError


if TYPE_CHECKING:
    from cherno.actor import ChernoCommandType


__all__ = ["apply_axes_correction"]


async def apply_axes_correction(
    command: ChernoCommandType,
    radec: tuple[float, float] | numpy.ndarray | None = None,
    rot: float | None = None,
    k_radec: float | None = None,
    k_rot: float | None = None,
):
    """Send corrections to the TCC. Corrections here are in arcsec."""

    assert command.actor

    state = command.actor.state

    guide_loop = state.guide_loop
    enabled_axes = state.enabled_axes

    # Correction applied in ra, dec, rot, scale, in arcsec.
    correction_applied = [0.0, 0.0, 0.0, 0.0]

    if radec is not None and "radec" in enabled_axes:
        corr_radec = numpy.array(radec) / 3600.0  # In degrees!

        default_k_radec: float = guide_loop["radec"]["pid"]["k"]
        k_radec = k_radec or default_k_radec

        min_corr_arcsec = guide_loop["radec"]["min_correction"]
        max_corr_arcsec = guide_loop["radec"]["max_correction"]

        if numpy.all(numpy.abs(corr_radec) < (min_corr_arcsec / 3600.0)):
            # Small correction. Do not apply.
            command.debug("Ignoring small ra/dec correction.")

        elif numpy.any(numpy.abs(corr_radec) > (max_corr_arcsec / 3600)):
            raise ChernoError("RA/Dec correction too large. Not applying correction.")

        else:
            corr_radec *= k_radec

            tcc_offset_cmd = await command.send_command(
                "tcc",
                f"offset arc {corr_radec[0]}, {corr_radec[1]} /computed",
            )

            if tcc_offset_cmd.status.did_fail:
                command.error("Failed applying RA/Dec correction.")
                return correction_applied

            correction_applied[0] = numpy.round(corr_radec[0] * 3600.0, 3)
            correction_applied[1] = numpy.round(corr_radec[1] * 3600.0, 3)

    if rot is not None and "rot" in enabled_axes:
        corr_rot = numpy.array(rot) / 3600.0  # In degrees!

        default_k_rot: float = guide_loop["rot"]["pid"]["k"]
        k_rot = k_rot or default_k_rot

        min_corr_arcsec = guide_loop["rot"]["min_correction"]
        max_corr_arcsec = guide_loop["rot"]["max_correction"]

        if numpy.all(numpy.abs(corr_rot) < (min_corr_arcsec / 3600.0)):
            # Small correction. Do not apply.
            command.debug("Ignoring small rotator correction.")

        elif numpy.any(numpy.abs(corr_rot) > (max_corr_arcsec / 3600)):
            raise ChernoError("Rotator correction too large. Not applying correction.")

        else:
            corr_rot *= k_rot

            tcc_offset_cmd = await command.send_command(
                "tcc",
                f"offset guide 0.0, 0.0, {corr_rot} /computed",
            )

            if tcc_offset_cmd.status.did_fail:
                command.error("Failed applying rotator correction.")
                return correction_applied

            correction_applied[2] = numpy.round(corr_rot * 3600.0, 3)

    return correction_applied


async def apply_focus_correction(
    command: ChernoCommandType,
    focus_corr: float,
    k_focus: float | None = None,
):
    """Send focus corrections to the TCC. Corrections are in microns."""

    assert command.actor

    state = command.actor.state

    guide_loop = state.guide_loop
    enabled_axes = state.enabled_axes

    min_corr = guide_loop["focus"]["min_correction"]
    max_corr = guide_loop["focus"]["max_correction"]

    if "focus" not in enabled_axes:
        return

    if numpy.abs(focus_corr) < min_corr:
        command.debug("Ignoring small focus correction.")
        return

    if numpy.abs(focus_corr) > max_corr:
        command.warning("Ignoring large focus correction.")
        return

    default_k_focus: float = guide_loop["focus"]["pid"]["k"]
    k_focus = k_focus or default_k_focus

    focus_corr = focus_corr * k_focus

    tcc_offset_cmd = await command.send_command(
        "tcc",
        f"set focus={focus_corr} /incremental",
    )

    if tcc_offset_cmd.status.did_fail:
        command.error("Failed applying focus correction.")
        return

    return numpy.round(focus_corr, 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-19
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from . import ChernoCommandType, cherno_parser


__all__ = ["status"]


@cherno_parser.command()
async def status(command: ChernoCommandType):
    """Reports the status of the system."""

    command.info(guider_status=hex(command.actor.state.status.value))
    command.info(enabled_axes=command.actor.state.enabled_axes)
    command.info(offset=command.actor.state.offset)

    for axis in ["radec", "rot", "focus"]:
        command.info(
            message={f"pid_{axis}": [command.actor.state.guide_loop[axis]["pid"]["k"]]}
        )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-09-13
# @Filename: astrometry.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import subprocess
import time

from typing import NamedTuple, Optional

from cherno.extraction import PathLike


class TimedProcess(NamedTuple):
    """A completed process which includes its elapsed time."""

    process: asyncio.subprocess.Process
    elapsed: float


class AstrometryNet:
    """A wrapper for the astrometry.net ``solve-field`` command.

    Parameters
    ----------
    configure_params
        Parameters to be passed to `.configure`.
    """

    def __init__(self, **configure_params):
        solve_field_cmd = subprocess.run(
            "which solve-field",
            shell=True,
            capture_output=True,
        )
        solve_field_cmd.check_returncode()

        self.solve_field_cmd = solve_field_cmd.stdout.decode().strip()

        self._options = {}
        self.configure(**configure_params)

    def configure(
        self,
        backend_config: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        sort_column: Optional[str] = None,
        sort_ascending: Optional[bool] = None,
        no_plots: Optional[bool] = None,
        ra: Optional[float] = None,
        dec: Optional[float] = None,
        radius: Optional[float] = None,
        scale_low: Optional[float] = None,
        scale_high: Optional[float] = None,
        scale_units: Optional[str] = None,
        dir: Optional[str] = None,
        **kwargs,
    ):
        """Configures how to run of ``solve-field```.

        The parameters this method accepts are identical to those of
        ``solve-field`` and are passed unchanged.

        Parameters
        ----------
        backend_config
            Use this config file for the ``astrometry-engine`` program.
        width
            Specify the field width, in pixels.
        height
            Specify the field height, in pixels.
        sort_column
            The FITS column that should be used to sort the sources.
        sort_ascending
            Sort in ascending order (smallest first);
            default is descending order.
        no_plot
            Do not produce plots.
        ra
            RA of field center for search, in degrees.
        dec
            Dec of field center for search, in degrees.
        radius
            Only search in indexes within ``radius`` degrees of the field
            center given by ``ra`` and ``dec``.
        scale_low
            Lower bound of image scale estimate.
        scale_high
            Upper bound of image scale estimate.
        scale_units
            In what units are the lower and upper bounds? Choices:
            ``'degwidth'``, ``'arcminwidth'``, ``'arcsecperpix'``,
            ``'focalmm'``.
        dir
            Path to the directory where all output files will be saved.
        """

        self._options = {
            "backend-config": backend_config,
            "width": width,
            "height": height,
            "sort-column": sort_column,
            "sort-ascending": sort_ascending,
            "no-plots": no_plots,
            "ra": ra,
            "dec": dec,
            "radius": radius,
            "scale-low": scale_low,
            "scale-high": scale_high,
            "scale-units": scale_units,
            "dir": dir,
            "overwrite": True,
        }
        self._options.update(kwargs)

        return

    def _build_command(self, files, options=None):
        """Builds the ``solve-field`` command to run."""

        if options is None:
            options = self._options

        flags = ["no-plots", "sort-ascending", "overwrite"]

        cmd = [self.solve_field_cmd]

        for option in options:
            if options[option] is None:
                continue
            if option in flags:
                if options[option] is True:
                    cmd.append("--" + option)
            else:
                cmd.append("--" + option)
                cmd.append(str(options[option]))

        cmd += list(files)

        return cmd

    async def run(
        self,
        files: PathLike | list[PathLike],
        shell: bool = True,
        stdout: Optional[PathLike] = None,
        stderr: Optional[PathLike] = None,
        **kwargs,
    ) -> TimedProcess:
        """Runs astrometry.net.

        Parameters
        ----------
        files
            List of files to be processed.
        shell
            Whether to call `subprocess.run` with ``shell=True``.
        stdout
            Path where to save the stdout output.
        stderr
            Path where to save the stderr output.
        kwargs
            Configuration parameters (see `.configure`) to override. The
            configuration applies only to this run of ``solve-field`` and it
            is not saved.

        Returns
        -------
        `subprocess.CompletedProcess`
            The completed process.

        """

        options = self._options.copy()
        options.update(kwargs)

        if not isinstance(files, (tuple, list)):
            files = [files]

        t0 = time.time()

        args = self._build_command(list(map(str, files)), options=options)

        if shell:
            cmd = await asyncio.create_subprocess_shell(
                " ".join(args),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        else:
            cmd = await asyncio.create_subprocess_exec(
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        stdout_bytes, stderr_bytes = await cmd.communicate()

        elapsed = time.time() - t0

        if stdout:
            with open(stdout, "wb") as out:
                out.write(" ".join(args).encode() + b"\n")
                out.write(stdout_bytes)

        if stderr:
            with open(stderr, "wb") as err:
                err.write(stderr_bytes)

        return TimedProcess(cmd, elapsed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-18
# @Filename: mock.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from . import ChernoCommandType, cherno_parser


__all__ = ["mock_command"]


@cherno_parser.command()
async def mock_command(command: ChernoCommandType):
    """Mocks some output. For developer use only."""

    command.info(fwhm_camera=["gfa4", 290, 2.315, 13, 10])
    command.info(fwhm_camera=["gfa2", 290, 2.566, 15, 14])
    command.info(fwhm_camera=["gfa4", 290, 2.315, 13, 10])
    command.info(fwhm_camera=["gfa3", 290, 2.409, 10, 6])
    command.info(fwhm_camera=["gfa5", 290, 2.2, 13, 8])
    command.info(focus_fit=[290, 2.198, 2.298e-06, -0.001441, 6.636, 0.891, -62.7])

    command.info(fwhm_camera=["gfa4", 291, 2.315, 13, 10])
    command.info(fwhm_camera=["gfa2", 291, 2.566, 15, 14])
    command.info(fwhm_camera=["gfa4", 291, 2.315, 13, 10])
    command.info(fwhm_camera=["gfa3", 291, 2.409, 10, 6])
    command.info(fwhm_camera=["gfa5", 291, 2.2, 13, 8])
    command.info(focus_fit=[291, 2.198, 2.298e-06, -0.001441, 6.636, 0.891, -62.7])

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-08
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import concurrent.futures
import pathlib
import warnings
from functools import partial

from typing import TYPE_CHECKING, Any, cast

import matplotlib.pyplot as plt
import numpy
import seaborn
from coordio.defaults import GFA_PIXEL_SIZE, PLATE_SCALE
from coordio.exceptions import CoordIOUserWarning
from coordio.utils import radec2wokxy

from cherno import config
from cherno.coordinates import gfa_to_wok
from cherno.exceptions import ChernoError


if TYPE_CHECKING:
    from cherno.acquisition import AcquisitionData
    from cherno.extraction import ExtractionData


warnings.simplefilter("ignore", category=CoordIOUserWarning)


async def run_in_executor(fn, *args, catch_warnings=False, executor="thread", **kwargs):
    """Runs a function in an executor.

    In addition to streamlining the use of the executor, this function
    catches any warning issued during the execution and reissues them
    after the executor is done. This is important when using the
    actor log handler since inside the executor there is no loop that
    CLU can use to output the warnings.

    In general, note that the function must not try to do anything with
    the actor since they run on different loops.

    """

    fn = partial(fn, *args, **kwargs)

    if executor == "thread":
        executor = concurrent.futures.ThreadPoolExecutor
    elif executor == "process":
        executor = concurrent.futures.ProcessPoolExecutor
    else:
        raise ValueError("Invalid executor name.")

    if catch_warnings:
        with warnings.catch_warnings(record=True) as records:
            with executor() as pool:
                result = await asyncio.get_event_loop().run_in_executor(pool, fn)

        for ww in records:
            warnings.warn(ww.message, ww.category)

    else:
        with executor() as pool:
            result = await asyncio.get_running_loop().run_in_executor(pool, fn)

    return result


def umeyama(X, Y):
    """Rigid alignment of two sets of points in k-dimensional Euclidean space.

    Given two sets of points in correspondence, this function computes the
    scaling, rotation, and translation that define the transform TR that
    minimizes the sum of squared errors between TR(X) and its corresponding
    points in Y.  This routine takes O(n k^3)-time.

    Parameters
    ----------
    X
        A ``k x n`` matrix whose columns are points.
    Y
        A ``k x n`` matrix whose columns are points that correspond to the
        points in X

    Returns
    -------
    c,R,t
        The  scaling, rotation matrix, and translation vector defining the
        linear map TR as ::

            TR(x) = c * R * x + t

        such that the average norm of ``TR(X(:, i) - Y(:, i))`` is
        minimized.

    Copyright: Carlo Nicolini, 2013

    Code adapted from the Mark Paskin Matlab version from
    http://openslam.informatik.uni-freiburg.de/data/svn/tjtf/trunk/matlab/ralign.m

    See paper by Umeyama (1991)

    """

    m, n = X.shape

    mx = X.mean(1)
    my = Y.mean(1)

    Xc = X - numpy.tile(mx, (n, 1)).T
    Yc = Y - numpy.tile(my, (n, 1)).T

    sx = numpy.mean(numpy.sum(Xc * Xc, 0))

    Sxy = numpy.dot(Yc, Xc.T) / n

    U, D, V = numpy.linalg.svd(Sxy, full_matrices=True, compute_uv=True)
    V = V.T.copy()

    r = numpy.linalg.matrix_rank(Sxy)
    S = numpy.eye(m)

    if r < (m - 1):
        raise ValueError("Not enough independent measurements")

    if numpy.linalg.det(Sxy) < 0:
        S[-1, -1] = -1
    elif r == m - 1:
        if numpy.linalg.det(U) * numpy.linalg.det(V) < 0:
            S[-1, -1] = -1

    R = numpy.dot(numpy.dot(U, S), V.T)
    c = numpy.trace(numpy.dot(numpy.diag(D), S)) / sx
    t = my - c * numpy.dot(R, mx)

    return c, R, t


def astrometry_fit(
    data: list[AcquisitionData],
    grid=(10, 10),
    offset: tuple | list = (0.0, 0.0, 0.0),
    obstime: float | None = None,
    scale_rms: bool = False,
):
    """Fits translation, rotation, and scale from a WCS solution."""

    offset_ra, offset_dec, offset_pa = offset

    xwok_gfa: list[float] = []
    ywok_gfa: list[float] = []
    xwok_astro: list[float] = []
    ywok_astro: list[float] = []

    default_offset = config.get("default_offset", (0.0, 0.0, 0.0))

    for d in data:

        camera_id = int(d.camera[-1])
        xidx, yidx = numpy.meshgrid(
            numpy.linspace(0, 2048, grid[0]),
            numpy.linspace(0, 2048, grid[1]),
        )
        xidx = xidx.flatten()
        yidx = yidx.flatten()

        coords: Any = d.wcs.pixel_to_world(xidx, yidx)
        ra = coords.ra.value
        dec = coords.dec.value

        for x, y in zip(xidx, yidx):
            xw, yw, _ = gfa_to_wok(x, y, camera_id)
            xwok_gfa.append(cast(float, xw))
            ywok_gfa.append(cast(float, yw))

        cos_dec = numpy.cos(numpy.deg2rad(d.field_dec))
        offset_ra_deg = offset_ra / cos_dec / 3600.0
        default_offset_ra_deg = default_offset[0] / cos_dec / 3600.0

        _xwok_astro, _ywok_astro, *_ = radec2wokxy(
            ra,
            dec,
            None,
            "GFA",
            d.field_ra - offset_ra_deg - default_offset_ra_deg,
            d.field_dec - offset_dec / 3600.0 - default_offset[1] / 3600.0,
            d.field_pa - offset_pa / 3600.0 - default_offset[2] / 3600.0,
            d.observatory.upper(),
            obstime,
        )

        xwok_astro += _xwok_astro.tolist()
        ywok_astro += _ywok_astro.tolist()

    X = numpy.array([xwok_gfa, ywok_gfa])
    Y = numpy.array([xwok_astro, ywok_astro])
    try:
        c, R, t = umeyama(X, Y)
    except ValueError:
        return False

    plate_scale = PLATE_SCALE[data[0].observatory]

    # delta_x and delta_y only align with RA/Dec if PA=0. Otherwise we need to
    # project using the PA.
    pa_rad = numpy.deg2rad(data[0].field_pa)
    delta_ra = t[0] * numpy.cos(pa_rad) + t[1] * numpy.sin(pa_rad)
    delta_dec = -t[0] * numpy.sin(pa_rad) + t[1] * numpy.cos(pa_rad)

    # Convert to arcsec and round up
    delta_ra = numpy.round(delta_ra / plate_scale * 3600.0, 3)
    delta_dec = numpy.round(delta_dec / plate_scale * 3600.0, 3)

    delta_rot = numpy.round(-numpy.rad2deg(numpy.arctan2(R[1, 0], R[0, 0])) * 3600.0, 1)
    delta_scale = numpy.round(c, 6)

    if scale_rms:
        xwok_astro /= delta_scale
        ywok_astro /= delta_scale

    delta_x = (numpy.array(xwok_gfa) - numpy.array(xwok_astro)) ** 2  # type: ignore
    delta_y = (numpy.array(ywok_gfa) - numpy.array(ywok_astro)) ** 2  # type: ignore

    xrms = numpy.sqrt(numpy.sum(delta_x) / len(delta_x))
    yrms = numpy.sqrt(numpy.sum(delta_y) / len(delta_y))
    rms = numpy.sqrt(numpy.sum(delta_x + delta_y) / len(delta_x))

    # Convert to arcsec and round up
    xrms = numpy.round(xrms / plate_scale * 3600.0, 3)
    yrms = numpy.round(yrms / plate_scale * 3600.0, 3)
    rms = numpy.round(rms / plate_scale * 3600.0, 3)

    return (delta_ra, delta_dec, delta_rot, delta_scale, xrms, yrms, rms)


def focus_fit(
    e_data: list[ExtractionData],
    plot: pathlib.Path | str | bool | None = None,
) -> tuple[float, float, float, float, float, float]:
    """Determines the optimal focus.

    Performs a least-squares polynomial fit to the focus data using
    a quadratic polynomial. Also calculates the r2 coefficient.

    Parameters
    ----------
    e_data
        Extraction data. Must include the observatory, FWHM, focus offset for each
        camera, and camera name.
    plot
        The path where to save the generated plot. If `None`, does not generate
        a plot.

    Returns
    -------
    results
        A tuple with the fitted FWHM, offset to the optimal focus in microns,
        coefficients of the fitted quadratic polynomial, and r2 coefficient.

    """

    def f(x, a, b, c):
        return a * x**2 + b * x + c

    cam = []
    x = []
    y = []
    weights = []

    observatory = e_data[0].observatory
    fwhm_to_microns = config["pixel_scale"][observatory] * GFA_PIXEL_SIZE

    for e_d in e_data:
        valid = e_d.regions.loc[e_d.regions.valid == 1]
        if len(valid) == 0 or e_d.fwhm_median < 0:
            continue

        cam += [int(e_d.camera[-1])] * len(valid)
        x += [e_d.focus_offset] * len(valid)

        # Convert FWHM to microns so that they are the same units as the focus offset.
        fwhm_microns = valid.fwhm * fwhm_to_microns
        y += fwhm_microns.values.tolist()

        # Calculate the weights as the inverse variance of the FWHM measurement.
        # Also add a subjective estimation of how reliable each camera is.
        ivar = 1 / (e_d.regions.loc[e_d.regions.valid == 1, "residual_fit"] ** 2)
        ivar_camera = config["cameras"]["focus_weight"][e_d.camera] * ivar
        weights += ivar_camera.values.tolist()

    cam = numpy.array(cam)
    x = numpy.array(x)
    y = numpy.array(y)
    weights = numpy.array(weights)
    weights = weights / weights.max()

    if len(numpy.unique(x)) < 3:
        raise ChernoError("Not enough data points to fit focus.")

    # Perform polynomial fit.
    a, b, c = numpy.polyfit(x, y, 2, w=weights, full=False)

    # Calculate the focus of the parabola and the associated FWHM.
    x_min = -b / 2 / a
    fwhm_fit = (a * x_min**2 + x_min * b + c) / fwhm_to_microns

    # Determine the r2 coefficient. See https://bit.ly/3LmH76j
    f_i = a * x**2 + b * x + c
    ybar = numpy.sum(y) / len(y)
    ss_res = numpy.sum((y - f_i) ** 2)
    ss_tot = numpy.sum((y - ybar) ** 2)
    r2 = 1 - ss_res / ss_tot

    if plot is not None:

        if isinstance(plot, (str, pathlib.Path)):
            outpath = pathlib.Path(plot)
        else:
            path = e_data[0].path.parent
            mjd = e_data[0].mjd
            seq = e_data[0].exposure_no
            outpath = path / "focus" / f"focus-{mjd}-{seq}.pdf"

        outpath.parent.mkdir(exist_ok=True, parents=True)

        seaborn.set_theme(style="darkgrid", palette="dark")

        with plt.ioff():  # type: ignore
            fig, ax = plt.subplots()

            for icam in sorted(numpy.unique(cam)):
                x_cam = x[cam == icam]
                y_cam = y[cam == icam]
                ax.scatter(
                    x_cam,
                    y_cam / fwhm_to_microns,
                    marker="o",  # type: ignore
                    edgecolor="None",
                    s=5,
                    label=f"GFA{icam}",
                )

            x0 = numpy.min(x)
            x1 = numpy.max(x)
            xs = numpy.linspace(x0 - 0.1 * (x1 - x0), x1 + 0.1 * (x1 - x0))
            ys = (a * xs**2 + b * xs + c) / fwhm_to_microns
            ax.plot(
                xs,
                ys,
                "r-",
                lw=2.0,
                label="Fitted polynomial",
            )

            ax.axvline(
                x=x_min,
                ls="--",
                c="b",
                lw=0.5,
                label="Focus offset (" + str(numpy.round(x_min, 1)) + r"$\rm \,\mu m$)",
            )

            ax.legend()

            ax.set_xlabel("Focus offset [microns]")
            ax.set_ylabel("FWHM [arcsec]")
            ax.set_title(
                f"MJD: {e_data[0].mjd} - Exp No: {e_data[0].exposure_no} - "
                f"FWHM: {numpy.round(fwhm_fit, 3)}  arcsec"
            )

            fig.savefig(str(outpath))

        seaborn.reset_orig()
        plt.close("all")

    return fwhm_fit, x_min, a, b, c, r2



# encoding: utf-8

from typing import cast

from sdsstools import get_logger, get_package_version
from sdsstools.configuration import __ENVVARS__, get_config


NAME = "sdss-cherno"

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(path).
log = get_logger(NAME)


__ENVVARS__["OBSERVATORY"] = "?"
config = get_config("cherno")


__version__ = cast(str, get_package_version(path=__file__, package_name=NAME))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-14
# @Filename: set.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import click

from . import ChernoCommandType, cherno_parser


__all__ = ["set"]


@cherno_parser.command()
@click.argument("OPTIONS", type=str, nargs=-1)
async def set(command: ChernoCommandType, options: tuple[str, ...]):
    """Sets guiding parameters.

    Valid parameters are:
        exposure-time [EXPTIME]
        pid radec|rot k [VALUE]
        axes [radec rot off]
        cameras CAMERAS

    """

    assert command.actor

    if len(options) == 0:
        return command.fail("No options passed")

    elif options[0] == "exposure-time":
        if len(options) < 2:
            return command.fail("Invalid number of parameters")

        exp_time = float(options[1])
        command.actor.state.exposure_time = exp_time

    elif options[0] == "pid":
        if len(options) < 4:
            return command.fail("Invalid number of parameters")

        axis = options[1]
        component = options[2]
        value = float(options[3])

        if axis not in ["radec", "rot", "focus"]:
            return command.fail(f"Invalid axis {axis}.")

        if value <= 0 or value > 1:
            return command.fail("Invalid value. Must be between 0 and 1.")

        command.info(message={f"pid_{axis}": [value]})
        command.actor.state.guide_loop[axis]["pid"][component] = value

    elif options[0] == "axes":

        if len(options) == 1 or (len(options) == 2 and options[1] in ["off", "none"]):
            axes = []
        else:
            axes = options[1:]

        command.actor.state.enabled_axes = list(axes)
        command.info(enabled_axes=command.actor.state.enabled_axes)

    elif options[0] == "cameras":
        if len(options) < 2:
            return command.fail("Invalid number of parameters")

        cameras = options[1:]
        command.actor.state.enabled_cameras = list(cameras)

    else:
        return command.fail("Invalid parameter.")

    return command.finish()
